/*
 *  Open HR20
 *
 *  target:     ATmega169 @ 4 MHz in Honnywell Rondostat HR20E
 *
 *  compiler:   WinAVR-20071221
 *              avr-libc 1.6.0
 *              GCC 4.2.2
 *
 *  copyright:  2008 Jiri Dobry (jdobry-at-centrum-dot-cz)
 *
 *  license:    This program is free software; you can redistribute it and/or
 *              modify it under the terms of the GNU Library General Public
 *              License as published by the Free Software Foundation; either
 *              version 2 of the License, or (at your option) any later version.
 *
 *              This program is distributed in the hope that it will be useful,
 *              but WITHOUT ANY WARRANTY; without even the implied warranty of
 *              MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *              GNU General Public License for more details.
 *
 *              You should have received a copy of the GNU General Public License
 *              along with this program. If not, see http:*www.gnu.org/licenses
 */

/*!
 * \file       controller.c
 * \brief      Controller for temperature
 * \author     Jiri Dobry <jdobry-at-centrum-dot-cz>
 * \date       $Date$
 * $Rev$
 */

#include <stdint.h>
#include <stdlib.h>
#include <avr/pgmspace.h>

#include "config.h"
#include "main.h"
#include "com.h"
#include "../common/rtc.h"
#include "adc.h"
#include "eeprom.h"
#include "controller.h"
#include "keyboard.h"

// global Vars for default values: temperatures and speed
uint8_t CTL_temp_wanted=0;   // actual desired temperature
uint8_t CTL_temp_wanted_last=0xff;   // desired temperature value used for last PID control
uint8_t CTL_temp_auto=0;   // actual desired temperature by timer
bool CTL_mode_auto=true;   // actual desired temperature by timer
uint8_t CTL_mode_window = 0; // open window (0=closed, >0 open-timmer)
#if (HW_WINDOW_DETECTION)
	static uint8_t window_timer=AVERAGE_LEN+1;
#else
	uint16_t CTL_open_window_timeout;
#endif
int8_t CTL_interatorCredit;
uint8_t CTL_creditExpiration;
#if BOOST_CONTROLER_AFTER_CHANGE
	uint8_t PID_boost_timeout;  //boost timout in minutes
	uint8_t PID_boost;  //boost value, either valve max or min
#endif
static uint16_t PID_update_timeout=AVERAGE_LEN+1;   // timer to next PID controler action/first is 16 sec after statup
int8_t PID_force_update=AVERAGE_LEN+1;      // signed value, val<0 means disable force updates \todo rename
uint8_t valveHistory[VALVE_HISTORY_LEN];

static uint8_t pid_Controller(int16_t setPoint, int16_t processValue, uint8_t old_result, bool updateNow);

uint8_t CTL_error=0;
uint8_t CTL_error_jr=0;

#if (HW_WINDOW_DETECTION)
static void CTL_window_detection(void) {
	bool w;
	// PORTE |= _BV(PE2); // enable pull-up
	// nop();nop();
	w = ((PINE & _BV(PE2))!=0) && config.window_open_detection_enable;
	if (!w) PORTE &= ~_BV(PE2); // disable pullup for save energy

	if (CTL_mode_window != w) {
		if (window_timer==0) {
			CTL_mode_window = w;
			PID_force_update = 0;
			// kb_events |= KB_EVENT_UPDATE_LCD;
		} else {
			window_timer--;
			return;
		}
	}
	window_timer=(w)?(config.window_close_detection_delay):(config.window_open_detection_delay);
}
#else
static void CTL_window_detection(void) {
    uint8_t i = (ring_buf_temp_avgs_pos+AVGS_BUFFER_LEN
        -((CTL_mode_window!=0)?config.window_close_detection_time:config.window_open_detection_time)
        )%AVGS_BUFFER_LEN;
    int16_t min = 10000;
    int16_t max = 0;
    while (1) {
        int16_t x = ring_buf_temp_avgs[i];
        if (x!=0) { // startup condition
            if (x<min) min = x;
            if (x>max) max = x;
        }
        if (i==ring_buf_temp_avgs_pos) break;
        i=(i+1)%AVGS_BUFFER_LEN;
    }
    if ((temp_average-min) > (int16_t) config.window_close_detection_diff) {
        if (CTL_mode_window!=0) {
            CTL_mode_window=0;
            PID_force_update = 0;
			//kb_events |= KB_EVENT_UPDATE_LCD;
        }
    } else {
        if ((CTL_mode_window==0) && ((max-temp_average) > (int16_t) config.window_open_detection_diff)) {
            CTL_mode_window=config.window_open_timeout;
            PID_force_update = 0;
			//kb_events |= KB_EVENT_UPDATE_LCD;
        }
    }
}
#endif

/*!
 *******************************************************************************
 *  Controller update
 *  \note call it once per second
 *  \param minute_ch is true when minute is changed
 *  \returns valve position
 *
 ******************************************************************************/
void CTL_update(bool minute_ch) {
#if (HW_WINDOW_DETECTION)
	PORTE |= _BV(PE2); // enable pull-up
#endif

    if ( minute_ch || (CTL_temp_auto==0) ) {
        // minutes changed or we need return to timers
        uint8_t t=RTC_ActualTimerTemperature(!(CTL_temp_auto==0));
        if (t!=0) {
            CTL_temp_auto=t;
            if (CTL_mode_auto) {
                CTL_temp_wanted=CTL_temp_auto;
                if ((PID_force_update<0)&&(CTL_temp_wanted!=CTL_temp_wanted_last)) {
                    PID_force_update=0;
                }
            }
        }
    }
#if BOOST_CONTROLER_AFTER_CHANGE 
	if ( minute_ch && (PID_boost_timeout>0)) {
		PID_boost_timeout--;
		if (PID_boost_timeout==0) {
		  PID_force_update = 0;
		}
	}
#endif
#if 0 
	int16_t MaxTempOffset;
	int16_t adc_dummy;
	if ( minute_ch && (ADC_TempCompensationTimer>0)) {
		ADC_TempCompensationTimer--;
		MaxTempOffset  = (int16_t)( MAX(0, MIN(config.RoomTemp_Vpos_MaxOffset,(int16_t)(valve_wanted-config.valve_min-1)*10) ) ); //-1 is position where heater starts to heat
		MaxTempOffset *= (int16_t)(config.RoomTemp_Offset_Vopen) *10;  //*10 because offesttemp must be in 0,01C
		MaxTempOffset /= (int16_t)(config.RoomTemp_Vpos_MaxOffset);
		adc_dummy = temp_average-ADC_LastMinuteTemp;
		if (ADC_LastMinuteTemp) {
			if ( (ADC_TempDeltaSum < MaxTempOffset+adc_dummy) && ( adc_dummy > 0) )  { 
				ADC_TempDeltaSum+=adc_dummy;
				ADC_LastMinuteTemp=temp_average-adc_dummy; 
			} else if ( (ADC_TempDeltaSum > MaxTempOffset+adc_dummy) && ( adc_dummy < 0) )  {
				ADC_TempDeltaSum+=adc_dummy;
				ADC_LastMinuteTemp=temp_average-adc_dummy; 
			}
		}
		// if (ADC_TempCompensationTimer < ADC_TempCompensationSetTime) { // wait RoomTemp_HeaterDelay minutes till heater & termistor react
		//	ADC_RoomTempOffset += ((MaxTempOffset - RoomTempOffset) * (int16_t)(ADC_TempCompensationSetTime - ADC_TempCompensationTimer)) / (int16_t)ADC_TempCompensationSetTime /2;	
			// TempIst = measuredTemp + TempMinOffset(TempOffet when valve closed) + TempOffset(depending on valve position)
		// }
	}
#endif

    CTL_window_detection();
    if (PID_update_timeout>0) PID_update_timeout--;
    if (PID_force_update>0) { 
		PID_force_update--;
	} else if ((PID_update_timeout == 0)||(PID_force_update==0)) {
        uint8_t temp;
        if ((CTL_temp_wanted<TEMP_MIN) || mode_window()) {
            temp = TEMP_MIN;  // frost protection to TEMP_MIN
        } else {
            temp = CTL_temp_wanted;
        }
        bool updateNow=(temp!=CTL_temp_wanted_last);
		  if (updateNow||(PID_update_timeout == 0)) {
            PID_update_timeout = (config.PID_interval * 5); // new PID pooling
			uint8_t new_valve;
            if (temp>TEMP_MAX) {
                new_valve = config.valve_max;
            } else {
                new_valve = pid_Controller(calc_temp(temp),temp_average,valveHistory[0],updateNow);
            }
					CTL_temp_wanted_last=temp;
			{	
				int8_t i;
				#if (BLOCK_INTEGRATOR_AFTER_VALVE_CHANGE) || (TEMP_COMPENSATE_OPTION)
					if (valveHistory[0]!=new_valve) {  
				#endif
						#if BLOCK_INTEGRATOR_AFTER_VALVE_CHANGE
							CTL_integratorBlock=CTL_INTEGRATOR_BLOCK;       //block Integrator if valve moves
						#endif

						#if 0
							ADC_TempCompensationTimer = config.RoomTemp_CompensationTime;
							if (new_valve < valveHistory[0])	{		// if cooling = new valve lower than last valve
								ADC_TempCompensationTimer += config.RoomTemp_CompensationTime; // cooling needs twice the time of heating
							} 
							ADC_TempCompensationSetTime = ADC_TempCompensationTimer;
							if ( (valveHistory[0] <= (config.valve_min+1)) && (new_valve > (config.valve_min +1)) ) {	// if heating and valve was closed before
								ADC_TempCompensationTimer += config.RoomTemp_HeaterDelay;  //time until heater starts changing temp
							}
						#endif
				#if (BLOCK_INTEGRATOR_AFTER_VALVE_CHANGE) || (TEMP_COMPENSATE_OPTION) 
				 	}
			    #endif

				for(i=VALVE_HISTORY_LEN-1; i>0; i--) {
					if (updateNow || (new_valve <= config.valve_max) || (new_valve >= config.valve_min)) {
						// condition inside loop is stupid, but produce shorter code
						valveHistory[i]=new_valve;
					} else  {
						valveHistory[i]=valveHistory[i-1];
					}
				}
				valveHistory[0]=new_valve;
			}
        }
        COM_print_debug(0);
        PID_force_update = -1; // invalid value = not used
    }
    // batt error detection
    if (bat_average) {
		if (bat_average < 20*(uint16_t)config.bat_low_thld) {
    	    CTL_error |=  CTL_ERR_BATT_LOW | CTL_ERR_BATT_WARNING;
	    } else {
	        if (bat_average < 20*(uint16_t)config.bat_warning_thld) {
	            CTL_error |=  CTL_ERR_BATT_WARNING;
				#if (BATT_ERROR_REVERSIBLE)
	            	CTL_error &= ~CTL_ERR_BATT_LOW;
				#endif
	        } else {
				#if (BATT_ERROR_REVERSIBLE)
	            	CTL_error &= ~(CTL_ERR_BATT_WARNING|CTL_ERR_BATT_LOW);
				#endif
	        }
	    }
	}
}

/*!
 *******************************************************************************
 *  Change controller temperature (+-)
 *
 *  \param ch relative change
 ******************************************************************************/
void CTL_temp_change_inc (int8_t ch) {
    CTL_temp_wanted+=ch;
	if (CTL_temp_wanted<TEMP_MIN-1) {
		CTL_temp_wanted= TEMP_MIN-1;
	} else if (CTL_temp_wanted>TEMP_MAX+1) {
		CTL_temp_wanted= TEMP_MAX+1;
	}
	CTL_mode_window = 0;
    PID_force_update = 9;
}

static uint8_t menu_temp_rewoke;
/*!
 *******************************************************************************
 *  Change controller mode
 *
 ******************************************************************************/
void CTL_change_mode(int8_t m) {
    if (m == CTL_CHANGE_MODE) {
        // change
  		menu_temp_rewoke=CTL_temp_auto;
		#if BOOST_CONTROLER_AFTER_CHANGE
			PID_boost_timeout = 0; //to disable boost change mode, then change setpoint and change back mode
		#endif
        CTL_mode_auto=!CTL_mode_auto;
        PID_force_update = 9;
    } else if (m == CTL_CHANGE_MODE_REWOKE) {
        //rewoke
  		CTL_temp_auto=menu_temp_rewoke;
        CTL_mode_auto=!CTL_mode_auto;
        PID_force_update = 9;
    } else {
        if (m >= 0) CTL_mode_auto=m;
        PID_force_update = 0;
    }
    if (CTL_mode_auto && (m != CTL_CHANGE_MODE_REWOKE)) {
    	CTL_temp_wanted=(CTL_temp_auto=RTC_ActualTimerTemperature(false));
    	// CTL_temp_auto=0;  //refresh wanted temperature in next step
    }
    CTL_mode_window = 0;
}

//! Summation of errors, used for integrate calculations
int32_t sumError=0;
static uint8_t lastErrorSign = 0;
static uint16_t lastAbsError = 0xFF;  //set invalid
static uint16_t last2AbsError= 0xFE;  //set invalid and different to lastAbsError
uint8_t CTL_integratorBlock;
//! The scalling_factor for PID constants
#define scalling_factor  (256)
#if (scalling_factor != 256)
    #error optimized only for (scalling_factor == 256)
#endif

//! Last process value, used to find derivative of process value. 	 
static int16_t lastProcessValue=0;
/*! \brief non-linear  PID control algorithm.
 *
 *  Calculates output from setpoint, process value and PID status.
 *
 *  \param setPoint  Desired value.
 *  \param processValue  Measured value.
 */

static uint16_t lastTempChangeErrorAbs=0xffff;
int32_t lastTempChangeSumError=0;
uint8_t average_sumerr_count=0; // counting # of averages, just for diagnosis purposes

static void testIntegratorRevert(uint16_t absErr) {
	 if ((absErr>=((lastTempChangeErrorAbs*3)>>2)) 
		&& (absErr > (I_ERR_TOLLERANCE_AROUND_0*2))) {
		// if error could not be reduced to 3/4 and Error is larger than I_ERR_TOLLERANCE_AROUND_0*2 C
		sumError=lastTempChangeSumError;
	}
	lastTempChangeErrorAbs = 0xffff; // function can be called more time but only first is valid
}


static uint8_t pid_Controller(int16_t setPoint, int16_t processValue, uint8_t old_result, bool updateNow)
{
  int32_t /*error2,*/ pi_term;
  int16_t error16;
#if TEMP_COMPENSATE_OPTION 
	// int16_t MaxTempOffset;
	int16_t adc_dummy1;
	int16_t adc_dummy2;
	// if ( minute_ch && (ADC_TempCompensationTimer>0)) {
	//	ADC_TempCompensationTimer--;
	//		MaxTempOffset  = (int16_t)( MAX(0, MIN(config.RoomTemp_Vpos_MaxOffset,(int16_t)(valve_wanted-config.valve_min-1)*10) ) ); //-1 is position where heater starts to heat
	//		MaxTempOffset *= (int16_t)(config.RoomTemp_Offset_Vopen) ;  //*10 because offesttemp must be in 0,01C
	//		MaxTempOffset /= (int16_t)(config.RoomTemp_Vpos_MaxOffset);
		adc_dummy1 = MIN( (int16_t)config.RoomTemp_Vpos_MaxOffset *10, MAX(0,(int16_t)(valve_wanted-config.valve_min) -1) *100 ); //-1 is position where heater starts to heat
		adc_dummy1 -= ADC_RoomTemp_ValveAverage;
		if (adc_dummy1 > 0) {
			 ADC_RoomTemp_ValveAverage += adc_dummy1 / (int16_t)config.RoomTemp_OffsetIncrease;
		} else {
			 ADC_RoomTemp_ValveAverage += adc_dummy1 / (int16_t)config.RoomTemp_OffsetDecrease;
		}
		
		adc_dummy2 = (ADC_RoomTemp_ValveAverage * (int16_t)config.RoomTemp_Offset_Vopen) / (int16_t)config.RoomTemp_Vpos_MaxOffset /10; //valve average to temp offset
		
		adc_dummy2 = adc_dummy2 - ADC_RoomTemp_Offset;
		if ( adc_dummy2 > 0)  {  // actual offest smaller than wanted offset
			adc_dummy2 = MAX(0, MIN(processValue-lastProcessValue, adc_dummy2) );
		}
		ADC_RoomTemp_Offset += adc_dummy2;
		processValue += adc_dummy2;

		// if (ADC_LastMinuteTemp) {
		//	adc_dummy2 = 0;
		//	if ( ADC_RoomTemp_WantedOffset > MaxTempOffset)  {  // larger than expected offset
		//		ADC_RoomTemp_WantedOffset -= ((config.RoomTemp_Vpos_MaxOffset-MaxTempOffset) *config.RoomTemp_OffsetDecrease) /config.RoomTemp_Vpos_MaxOffset/10;
		//		// ADC_LastMinuteTemp=temp_average-adc_dummy; 
		//	} else if (ADC_RoomTemp_WantedOffset < MaxTempOffset)  {
		//		ADC_RoomTemp_WantedOffset += (MaxTempOffset * config.RoomTemp_OffsetIncrease) / config.RoomTemp_Vpos_MaxOffset /10;
		//	}
			// } else if ( (ADC_RoomTemp_WantedOffset < MaxTempOffset+adc_dummy) && ( adc_dummy > 0) )  {
		// }
		// if (ADC_TempCompensationTimer < ADC_TempCompensationSetTime) { // wait RoomTemp_HeaterDelay minutes till heater & termistor react
		//	ADC_RoomTempOffset += ((MaxTempOffset - RoomTempOffset) * (int16_t)(ADC_TempCompensationSetTime - ADC_TempCompensationTimer)) / (int16_t)ADC_TempCompensationSetTime /2;	
		// TempIst = measuredTemp + TempMinOffset(TempOffet when valve closed) + TempOffset(depending on valve position)
		// }
#endif

  error16 = setPoint - processValue;

  // maximum error is 12 degree C
  if (error16 > 1200) {
    error16=1200;
  } else if (error16 < -1200) {
    error16=-1200;
  }

  {
	  int16_t absErr = abs(error16);
	  if (updateNow) {
		  CTL_interatorCredit=config.I_max_credit;
		  CTL_creditExpiration=config.I_credit_expiration;
		  CTL_integratorBlock=CTL_INTEGRATOR_BLOCK; // do not allow update integrator immediately after temp change
		  testIntegratorRevert(lastAbsError);
		  if ((lastAbsError==last2AbsError) && (lastAbsError<(I_ERR_TOLLERANCE_AROUND_0*2))) {//stable error & error < 0,3C)
		  	average_sumerr_count++;  //just for diagnosis, how often do we average
			lastTempChangeSumError += (sumError-lastTempChangeSumError)/I_SUMERR_CHANGE_WEIGHT; 
			//imapct of SumErr changes on lastSummError, averaging for integrator revert
		  }
		  lastTempChangeErrorAbs = absErr;
		  // lastTempChangeSumError = sumError;
		  #if BOOST_CONTROLER_AFTER_CHANGE
		  	  PID_boost_timeout = 0;
			  if (absErr>=(int16_t)config.temp_boost_error) { // && CTL_temp_wanted_last<>TEMP_MIN  // error large enough to start boost (0,3C)
				if (error16 >= 0) {
					PID_boost = config.valve_max;
					PID_boost_timeout = config.temp_boost_time_heat;
				} else {
					PID_boost = config.valve_min;
					PID_boost_timeout = config.temp_boost_time_cool;
				}
				PID_boost_timeout = (uint8_t)(MIN(255,abs(error16)/10*(int16_t)PID_boost_timeout/CTL_BOOST_TEMPCHANGE)); //boosttime=error/10(0,1C)*time/CTL_BOOST_TEMPCHANGE(0,5C)
			  } 			  	
		  #endif
	  } else {
		if (CTL_integratorBlock == 0) {
			if (CTL_creditExpiration>0) {
				CTL_creditExpiration--;
			} else {
				CTL_interatorCredit=0;
			}
			if ((error16 >= 0) ? (old_result < config.valve_max) : (old_result > config.valve_min)) {
			  if (  // ((lastErrorSign != ((uint8_t)(error16>>8)&0x80))) ||  //sign of last error16 != sign of current
				((absErr==last2AbsError) && (absErr<=I_ERR_TOLLERANCE_AROUND_0))) {  //abserror around 0 with slow change
				  CTL_interatorCredit=config.I_max_credit; 
				  CTL_creditExpiration=config.I_credit_expiration;
				  goto INTEGRATOR; // next integration, do not change CTL_interatorCredit
			  }
			  if (CTL_interatorCredit>0) {
				  if (absErr >= last2AbsError) { // error can grow only limited time 
					  CTL_interatorCredit-=(absErr/I_ERR_WEIGHT)+1; // max is 1200/20+1 = 61
					  INTEGRATOR:
					  sumError += error16*8;
				  }
			  }
			} 
			if (CTL_interatorCredit<=0) {
				// credit is empty, test result
				testIntegratorRevert(absErr);
			} 
		} else {
			CTL_integratorBlock--;
		}
		last2AbsError = lastAbsError;
		lastAbsError = absErr;
	    lastErrorSign = (uint8_t)(error16>>8)&0x80;

      }
  }
  lastProcessValue = processValue;
  
  #if BOOST_CONTROLER_AFTER_CHANGE
	if (config.ValveManual) {
		return config.ValveManual;  //return manual valve override for test purposes
	}
	if (PID_boost_timeout > 0) {
		CTL_integratorBlock=CTL_INTEGRATOR_BLOCK;       //block Integrator
		CTL_creditExpiration=config.I_credit_expiration;
		return PID_boost;  //return valve_MAX or MIN
	}
  #endif
  
  if (config.I_Factor > 0) {
	  int32_t maxSumError;
      // for overload protection: maximum is scalling_factor*scalling_factor*50/1 = 3276800
      maxSumError = ((int32_t)scalling_factor*(int32_t)scalling_factor*50)/config.I_Factor;
	  if(sumError > maxSumError){
		sumError = maxSumError;
	  } else if(sumError < -maxSumError){
		sumError = -maxSumError;
	  }
  }

  pi_term = (int32_t)(error16);
  pi_term *=  pi_term * (int32_t)config.P3_Factor;
  pi_term >>= 8;
  pi_term += ((uint16_t)config.P_Factor <<8);
  pi_term *= (int32_t)error16;
  pi_term += (int32_t)(config.I_Factor) * sumError; // maximum is 65536*50=(scalling_factor*scalling_factor*50/I_Factor)*I_Factor
  /* 
   * pi_term - > for overload limit: 
   * maximum is +-(((255*1200*1200/256)+255)*1200+65536*50)
   * = +-1724832800 fit into signed 32bit
  */
  pi_term += (int32_t)(config.valve_center)*scalling_factor*scalling_factor;
  pi_term >>= 8; // /=scalling_factor

  if(pi_term > (int32_t)((uint16_t)config.valve_max*scalling_factor)){
    return config.valve_max;
  } else if(pi_term < 0){
    return config.valve_min;
  }
  // now we can use 16bit value ( 0 < pi_term < 25600 )
  {
    uint16_t pi_term16 = pi_term;

	{
		bool gt = (uint8_t)(pi_term16>>8 /*/scalling_factor*/) >= old_result;
		// asymetric round, ignore changes < 3/4%
		pi_term16 += scalling_factor/2; // prepare for round
		if (gt) {
			pi_term16 -= config.valve_hysteresis; // prepare for round
		} else {
			pi_term16 += config.valve_hysteresis; // prepare for round
		}
		pi_term16 >>= 8; // /= scalling_factor;
	}
	if(pi_term16 < config.valve_min) {
      return config.valve_min;
    }
    
	return (uint8_t)pi_term16;
  }
}

